/**
 * OW2 Util
 * Copyright (C) 2008 SERLI
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: ObrResourceFetcher.java 4392 2008-12-15 14:05:41Z alitokmen $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.plan.fetcher.impl.obr;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;

import org.apache.felix.bundlerepository.RepositoryAdminImpl;
import org.osgi.framework.BundleContext;
import org.osgi.service.obr.RepositoryAdmin;
import org.osgi.service.obr.Resource;
import org.ow2.util.file.FileUtils;
import org.ow2.util.file.FileUtilsException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.plan.bindings.deploymentplan.obr.ObrDeployment;
import org.ow2.util.plan.bindings.exceptions.InvalidDeploymentException;
import org.ow2.util.plan.bindings.repository.Repository;
import org.ow2.util.plan.bindings.repository.RepositoryKind;
import org.ow2.util.plan.fetcher.api.IResourceFetcher;
import org.ow2.util.plan.fetcher.api.exceptions.FetcherException;
import org.ow2.util.plan.fetcher.api.exceptions.ResourceFetcherNotResolvedException;
import org.ow2.util.plan.fetcher.impl.AbsResourceFetcherImpl;

/**
 * Implementation of the Resource Fetcher that can process resources on OBR repositories.
 * @author Mickaël LEDUQUE
 */
public class ObrResourceFetcher extends AbsResourceFetcherImpl implements IResourceFetcher {

    /**
     * The logger.
     */
    private static final Log logger = LogFactory.getLog(ObrResourceFetcher.class);

    /**
     * The local copy of the file.
     */
    private File localFile = null;

    /**
     * The bundle context.
     */
    private BundleContext bundleContext = null;

    /**
     * The URL where the resource was found.
     */
    private URL resourceURL = null;

    /**
     * The date of the last check.
     */
    private Long lastCheck = null;

    /**
     * {@inheritDoc}
     */
    public File getResource() throws ResourceFetcherNotResolvedException {
        if (this.localFile == null) {
            throw new ResourceFetcherNotResolvedException();
        }
        return this.localFile;
    }


    /**
     * {@inheritDoc}
     */
    public boolean hasChanged() throws FetcherException {
        URLConnection connection = null;

        try {
            connection = this.resourceURL.openConnection();
        } catch (IOException e) {
            throw new FetcherException(e);
        }
        long time = System.currentTimeMillis() - 1;
        long changed = connection.getLastModified();
        if (this.lastCheck < changed) {
            // The resource has changed
            return true;
        } else {
            this.lastCheck = time;
            return false;
        }
    }


    /**
     * {@inheritDoc}
     */
    public void resolve() throws FetcherException {
        if (this.deployment == null) {
            throw new FetcherException(new InvalidDeploymentException("null deployment"));
        }
        if (!(this.deployment instanceof ObrDeployment)) {
            throw new FetcherException(new InvalidDeploymentException("Bad deployment type: "
                    + this.deployment.getClass().getName()));
        }
        ObrDeployment obrDeployment = (ObrDeployment) this.deployment;
        if (this.bundleContext == null) {
            throw new FetcherException(
                    "Cannot resolve this OBR resource fetcher with a null bundle context.");
        }
        // Build and fill the repository admin.
        // TODO remove the default (oscar) OBR repository
        RepositoryAdmin repositoryAdmin = new RepositoryAdminImpl(this.bundleContext);
        for (Repository repository : this.getRepositoryManager()) {
            if (repository.getType() == RepositoryKind.OBR) {
                URL url = null;
                try {
                     url = new URL(repository.getUrl());
                } catch (MalformedURLException e) {
                    logger.warn("Bad url in repository {0}; exception: {1}", repository, e);
                }
                if (url != null) {
                    try {
                        repositoryAdmin.addRepository(url);
                    } catch (Exception e) {
                        logger.warn(
                                "Could not insert repository {0} into the OBR repository admin.",
                                url);
                    }
                }
            }
        }
        String filter = this.buildFilterString(obrDeployment);
        if (filter == null) {
            logger.error(
                    "OBR deployment {0} resolves in null search filter - aborting deployment",
                    obrDeployment);
            throw new FetcherException(
                    "Null search filter for deployment " + obrDeployment.toString());
        }
        Resource[] discoveryResult = repositoryAdmin.discoverResources(filter);
        if (discoveryResult.length == 0) {
            logger.error("OBR deployment {0} search returns no matching resource.",
                    obrDeployment);
            throw new FetcherException(
                    "No matching search result for deployment " + obrDeployment.toString());
        }
        if (discoveryResult.length > 1) {
            logger.error("OBR deployment {0} search returns more than one resource.",
                    obrDeployment);
            throw new FetcherException(
                    "Too many search results for deployment " + obrDeployment.toString());
        }
        Resource matchingResource = discoveryResult[0];

        /* We could use the OSGi infrastructure here, but this needs some changes to
         * the fetcher api - once again?
         * For example, with a usedAsBundle in the deployment, we could do
         * if (obrDeployment.isUsedAsFlag()) {
         *      Resolver resolver = repositoryAdmin.resolver();
         *      resolver.add(matchingResource);
         *      if (!resolver.resolve()) {
         *          logger.error("OBR deployment {0} could not be resolved.", obrDeployment);
         *          throw new FetcherException(
         *                  "Cannot resolve OBR deployment " + obrDeployment.toString());
         *      }
         *      resolver.deploy(false);
         * }
         * But then we need to get the resulting bundle context back and return
         * Archives instead of Files (thus the API change).
         * Some BundleArchive/Factory is also needed.
         */

        resourceURL = matchingResource.getURL();
        this.localFile = makeLocalFilePath(this.localRepositoriesBaseDir,
                                           // TODO get back the real repository name
                                           matchingResource.getRepository().getName(),
                                           matchingResource.getSymbolicName());
        //Create the local file path.
        if (localFile != null) {
            if (!localFile.getParentFile().exists()) {
                localFile.getParentFile().mkdirs();
            }
        }
        // Copy the data to the local file.
        this.update();

    }


    /**
     * {@inheritDoc}
     */
    public void update() throws FetcherException {
        this.lastCheck = System.currentTimeMillis();
        InputStream input = null;
        try {
            input = resourceURL.openStream();
        } catch (IOException e) {
            logger.error("Exception while opening connection to resource {0}", resourceURL);
            throw new FetcherException("Exception while opening connection", e);
        }
        try {
            FileUtils.dump(input, localFile);
        } catch (FileUtilsException e) {
            logger.error("Exception while copying data from resource {0};exception: {1}", resourceURL, e);
            throw new FetcherException("Exception while copying resource data", e);
        }

    }

    public void setBundleContext(final BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    public BundleContext getBundleContext() {
        return this.bundleContext;
    }

    /**
     * Builds an OSGi filter from the given deployment.
     * @param obrDeployment the deployment.
     * @return the filter.
     */
    private String buildFilterString(final ObrDeployment obrDeployment) {
        if (obrDeployment == null) {
            return null;
        }
        String symbolicName = obrDeployment.getBundleSymbolicName();
        String bundleVersion = obrDeployment.getBundleVersion();
        String filter = obrDeployment.getFilter();
        String requireService = obrDeployment.getRequireService();

        StringBuilder sb = new StringBuilder("(&");
        if (symbolicName != null) {
            sb.append("(symbolicname=");
            sb.append(symbolicName);
            sb.append(")");
        }
        if (bundleVersion != null) {
            sb.append("(version");
            // version if like =1.0.0 or >=1.2
            sb.append(bundleVersion);
            sb.append(")");
        }
        if (filter != null) {
            sb.append(filter);
        }
        if (requireService != null) {
            sb.append("(service=");
            sb.append(requireService);
            sb.append(")");
        }
        sb.append(")");

        return sb.toString();
    }

    /**
     * {@inheritDoc}
     */
    public RepositoryKind getSupportedRepositoryType() {
        return RepositoryKind.OBR;
    }
}
